@@ -1,6 +1,25 @@
-/** @file main.c
-    mcs51 specific general functions.
+/*-------------------------------------------------------------------------
+  main.c - mcs51 specific general functions
 
+  Copyright (C) 1998, Sandeep Dutta . sandeep.dutta@usa.net
+  Copyright (C) 1999, Jean-Louis VERN.jlvern@writeme.com
+  Copyright (C) 2000, Michael Hope
+
+  This program is free software; you can redistribute it and/or modify it
+  under the terms of the GNU General Public License as published by the
+  Free Software Foundation; either version 2, or (at your option) any
+  later version.
+
+  This program is distributed in the hope that it will be useful,
+  but WITHOUT ANY WARRANTY; without even the implied warranty of
+  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+  GNU General Public License for more details.
+
+  You should have received a copy of the GNU General Public License
+  along with this program; if not, write to the Free Software
+  Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+-------------------------------------------------------------------*/
+/*
     Note that mlh prepended _mcs51_ on the static functions.  Makes
     it easier to set a breakpoint using the debugger.
 */
@@ -8,6 +27,8 @@
 #include "main.h"
 #include "ralloc.h"
 #include "gen.h"
+#include "peep.h"
+#include "rtrack.h"
 #include "dbuf_string.h"
 #include "../SDCCutil.h"
 
@@ -16,6 +37,19 @@ static char _defaultRules[] =
 #include "peeph.rul"
 };
 
+#define OPTION_STACK_SIZE       "--stack-size"
+
+static OPTION _mcs51_options[] =
+  {
+    { 0, OPTION_STACK_SIZE,  &options.stack_size, "Tells the linker to allocate this space for stack", CLAT_INTEGER },
+    { 0, "--parms-in-bank1", &options.parms_in_bank1, "use Bank1 for parameter passing"},
+    { 0, "--pack-iram",      NULL, "Tells the linker to pack variables in internal ram (default)"},
+    { 0, "--no-pack-iram",   &options.no_pack_iram, "Deprecated: Tells the linker not to pack variables in internal ram"},
+    { 0, "--acall-ajmp",     &options.acall_ajmp, "Use acall/ajmp instead of lcall/ljmp" },
+    { 0, "--no-ret-without-call", &options.no_ret_without_call, "Do not use ret independent of acall/lcall" },
+    { 0, NULL }
+  };
+
 /* list of key words used by msc51 */
 static char *_mcs51_keywords[] =
 {
@@ -26,26 +60,21 @@ static char *_mcs51_keywords[] =
   "critical",
   "data",
   "far",
+  "generic",
   "idata",
   "interrupt",
+  "naked",
   "near",
+  "nonbanked",
+  "overlay",
   "pdata",
   "reentrant",
+  "sbit",
   "sfr",
   "sfr16",
   "sfr32",
-  "sbit",
   "using",
   "xdata",
-  "_data",
-  "_code",
-  "_generic",
-  "_near",
-  "_xdata",
-  "_pdata",
-  "_idata",
-  "_naked",
-  "_overlay",
   NULL
 };
 
@@ -63,7 +92,7 @@ _mcs51_init (void)
 }
 
 static void
-_mcs51_reset_regparm (void)
+_mcs51_reset_regparm (struct sym_link *funcType)
 {
   regParmFlg = 0;
   regBitParmFlg = 0;
@@ -72,38 +101,45 @@ _mcs51_reset_regparm (void)
 static int
 _mcs51_regparm (sym_link * l, bool reentrant)
 {
-    if (IS_SPEC(l) && (SPEC_NOUN(l) == V_BIT)) {
-        /* bit parameters go to b0 thru b7 */
-        if (reentrant && (regBitParmFlg < 8)) {
-            regBitParmFlg++;
-            return 12 + regBitParmFlg;
+  if (IS_SPEC(l) && (SPEC_NOUN(l) == V_BIT))
+    {
+      /* bit parameters go to b0 thru b7 */
+      if (reentrant && (regBitParmFlg < 8))
+        {
+          regBitParmFlg++;
+          return 12 + regBitParmFlg;
         }
-        return 0;
+      return 0;
     }
-    if (options.parms_in_bank1 == 0) {
-        /* simple can pass only the first parameter in a register */
-        if (regParmFlg)
-            return 0;
+  if (options.parms_in_bank1 == 0)
+    {
+      /* simple can pass only the first parameter in a register */
+      if (regParmFlg)
+        return 0;
 
-        regParmFlg = 1;
-        return 1;
-    } else {
-        int size = getSize(l);
-        int remain ;
+      regParmFlg = 1;
+      return 1;
+    }
+  else
+    {
+      int size = getSize(l);
+      int remain;
 
-        /* first one goes the usual way to DPTR */
-        if (regParmFlg == 0) {
-            regParmFlg += 4 ;
-            return 1;
+      /* first one goes the usual way to DPTR */
+      if (regParmFlg == 0)
+        {
+          regParmFlg += 4 ;
+          return 1;
         }
-        /* second one onwards goes to RB1_0 thru RB1_7 */
-        remain = regParmFlg - 4;
-        if (size > (8 - remain)) {
-            regParmFlg = 12 ;
-            return 0;
+      /* second one onwards goes to RB1_0 thru RB1_7 */
+      remain = regParmFlg - 4;
+      if (size > (8 - remain))
+        {
+          regParmFlg = 12 ;
+          return 0;
         }
-        regParmFlg += size ;
-        return regParmFlg - size + 1;
+      regParmFlg += size ;
+      return regParmFlg - size + 1;
     }
 }
 
@@ -119,9 +155,8 @@ _mcs51_parseOptions (int *pargc, char **argv, int *i)
 static void
 _mcs51_finaliseOptions (void)
 {
-  if (options.noXinitOpt) {
+  if (options.noXinitOpt)
     port->genXINIT=0;
-  }
 
   switch (options.model)
     {
@@ -136,6 +171,7 @@ _mcs51_finaliseOptions (void)
       port->s.gptr_size = 3;
       break;
     case MODEL_LARGE:
+    case MODEL_HUGE:
       port->mem.default_local_map = xdata;
       port->mem.default_globl_map = xdata;
       port->s.gptr_size = 3;
@@ -146,9 +182,19 @@ _mcs51_finaliseOptions (void)
       break;
     }
 
-  if (options.parms_in_bank1) {
-      addSet(&preArgvSet, Safe_strdup("-DSDCC_PARMS_IN_BANK1"));
-  }
+  if (options.parms_in_bank1)
+    addSet(&preArgvSet, Safe_strdup("-DSDCC_PARMS_IN_BANK1"));
+
+  /* mcs51 has an assembly coded float library that's almost always reentrant */
+  if (!options.useXstack)
+    options.float_rent = 1;
+
+  if (options.omitFramePtr)
+    port->stack.reent_overhead = 0;
+
+  /* set up external stack location if not explicitly specified */
+  if (!options.xstack_loc)
+    options.xstack_loc = options.xdata_loc;
 }
 
 static void
@@ -157,7 +203,7 @@ _mcs51_setDefaultOptions (void)
 }
 
 static const char *
-_mcs51_getRegName (struct regs *reg)
+_mcs51_getRegName (const struct reg_info *reg)
 {
   if (reg)
     return reg->name;
@@ -167,10 +213,11 @@ _mcs51_getRegName (struct regs *reg)
 static void
 _mcs51_genAssemblerPreamble (FILE * of)
 {
-    if (options.parms_in_bank1) {
-        int i ;
-        for (i=0; i < 8 ; i++ )
-            fprintf (of,"b1_%d = 0x%x \n",i,8+i);
+  if (options.parms_in_bank1)
+    {
+      int i;
+      for (i=0; i < 8 ; i++ )
+        fprintf (of, "\tb1_%d = 0x%x \n", i, 8+i);
     }
 }
 
@@ -180,16 +227,17 @@ _mcs51_genIVT (struct dbuf_s * oBuf, symbol ** interrupts, int maxInterrupts)
 {
   int i;
 
-  dbuf_printf (oBuf, "\tljmp\t__sdcc_gsinit_startup\n");
+  dbuf_printf (oBuf, "\t%cjmp\t__sdcc_gsinit_startup\n", options.acall_ajmp?'a':'l');
+  if((options.acall_ajmp)&&(maxInterrupts)) dbuf_printf (oBuf, "\t.ds\t1\n");
 
   /* now for the other interrupts */
   for (i = 0; i < maxInterrupts; i++)
     {
       if (interrupts[i])
         {
-          dbuf_printf (oBuf, "\tljmp\t%s\n", interrupts[i]->rname);
+          dbuf_printf (oBuf, "\t%cjmp\t%s\n", options.acall_ajmp?'a':'l', interrupts[i]->rname);
           if ( i != maxInterrupts - 1 )
-            dbuf_printf (oBuf, "\t.ds\t5\n");
+            dbuf_printf (oBuf, "\t.ds\t%d\n", options.acall_ajmp?6:5);
         }
       else
         {
@@ -241,7 +289,8 @@ _mcs51_genInitStartup (FILE *of)
 
 
 /* Generate code to copy XINIT to XISEG */
-static void _mcs51_genXINIT (FILE * of) {
+static void _mcs51_genXINIT (FILE * of)
+{
   tfprintf (of, "\t!global\n", "__mcs51_genXINIT");
 
   if (!getenv("SDCC_NOGENRAMCLEAR"))
@@ -252,23 +301,24 @@ static void _mcs51_genXINIT (FILE * of) {
 /* Do CSE estimation */
 static bool cseCostEstimation (iCode *ic, iCode *pdic)
 {
-    operand *result = IC_RESULT(ic);
-    sym_link *result_type = operandType(result);
+  operand *result = IC_RESULT(ic);
+  sym_link *result_type = operandType(result);
 
-    /* if it is a pointer then return ok for now */
-    if (IC_RESULT(ic) && IS_PTR(result_type)) return 1;
+  /* if it is a pointer then return ok for now */
+  if (IC_RESULT(ic) && IS_PTR(result_type)) return 1;
 
-    /* if bitwise | add & subtract then no since mcs51 is pretty good at it
-       so we will cse only if they are local (i.e. both ic & pdic belong to
-       the same basic block */
-    if (IS_BITWISE_OP(ic) || ic->op == '+' || ic->op == '-') {
-        /* then if they are the same Basic block then ok */
-        if (ic->eBBlockNum == pdic->eBBlockNum) return 1;
-        else return 0;
+  /* if bitwise | add & subtract then no since mcs51 is pretty good at it
+     so we will cse only if they are local (i.e. both ic & pdic belong to
+     the same basic block */
+  if (IS_BITWISE_OP(ic) || ic->op == '+' || ic->op == '-')
+    {
+      /* then if they are the same Basic block then ok */
+      if (ic->eBBlockNum == pdic->eBBlockNum) return 1;
+      else return 0;
     }
 
-    /* for others it is cheaper to do the cse */
-    return 1;
+  /* for others it is cheaper to do the cse */
+  return 1;
 }
 
 /* Indicate which extended bit operations this port supports */
@@ -277,7 +327,6 @@ hasExtBitOp (int op, int size)
 {
   if (op == RRC
       || op == RLC
-      || op == GETHBIT
       || op == GETABIT
       || op == GETBYTE
       || op == GETWORD
@@ -298,7 +347,11 @@ oclsExpense (struct memmap *oclass)
   return 0;
 }
 
-
+static bool
+_hasNativeMulFor (iCode *ic, sym_link *left, sym_link *right)
+{
+  return getSize (left) == 1 && getSize (right) == 1;
+}
 
 static int
 instructionSize(char *inst, char *op1, char *op2)
@@ -438,37 +491,37 @@ mcs51operanddata;
 
 static mcs51operanddata mcs51operandDataTable[] =
   {
-    {"a", A_IDX, -1},
-    {"ab", A_IDX, B_IDX},
-    {"ac", CND_IDX, -1},
-    {"acc", A_IDX, -1},
-    {"ar0", R0_IDX, -1},
-    {"ar1", R1_IDX, -1},
-    {"ar2", R2_IDX, -1},
-    {"ar3", R3_IDX, -1},
-    {"ar4", R4_IDX, -1},
-    {"ar5", R5_IDX, -1},
-    {"ar6", R6_IDX, -1},
-    {"ar7", R7_IDX, -1},
-    {"b", B_IDX, -1},
-    {"c", CND_IDX, -1},
-    {"cy", CND_IDX, -1},
-    {"dph", DPH_IDX, -1},
-    {"dpl", DPL_IDX, -1},
+    {"a",    A_IDX,   -1},
+    {"ab",   A_IDX,   B_IDX},
+    {"ac",   CND_IDX, -1},
+    {"acc",  A_IDX,   -1},
+    {"ar0",  R0_IDX,  -1},
+    {"ar1",  R1_IDX,  -1},
+    {"ar2",  R2_IDX,  -1},
+    {"ar3",  R3_IDX,  -1},
+    {"ar4",  R4_IDX,  -1},
+    {"ar5",  R5_IDX,  -1},
+    {"ar6",  R6_IDX,  -1},
+    {"ar7",  R7_IDX,  -1},
+    {"b",    B_IDX,   -1},
+    {"c",    CND_IDX, -1},
+    {"cy",   CND_IDX, -1},
+    {"dph",  DPH_IDX, -1},
+    {"dpl",  DPL_IDX, -1},
     {"dptr", DPL_IDX, DPH_IDX},
-    {"f0", CND_IDX, -1},
-    {"f1", CND_IDX, -1},
-    {"ov", CND_IDX, -1},
-    {"p", CND_IDX, -1},
-    {"psw", CND_IDX, -1},
-    {"r0", R0_IDX, -1},
-    {"r1", R1_IDX, -1},
-    {"r2", R2_IDX, -1},
-    {"r3", R3_IDX, -1},
-    {"r4", R4_IDX, -1},
-    {"r5", R5_IDX, -1},
-    {"r6", R6_IDX, -1},
-    {"r7", R7_IDX, -1},
+    {"f0",   CND_IDX, -1},
+    {"f1",   CND_IDX, -1},
+    {"ov",   CND_IDX, -1},
+    {"p",    CND_IDX, -1},
+    {"psw",  CND_IDX, -1},
+    {"r0",   R0_IDX,  -1},
+    {"r1",   R1_IDX,  -1},
+    {"r2",   R2_IDX,  -1},
+    {"r3",   R3_IDX,  -1},
+    {"r4",   R4_IDX,  -1},
+    {"r5",   R5_IDX,  -1},
+    {"r6",   R6_IDX,  -1},
+    {"r7",   R7_IDX,  -1},
   };
 
 static int
@@ -656,7 +709,7 @@ static int
 getInstructionSize (lineNode *line)
 {
   if (!line->aln)
-    line->aln = asmLineNodeFromLineNode (line);
+    line->aln = (asmLineNodeBase *) asmLineNodeFromLineNode (line);
 
   return line->aln->size;
 }
@@ -665,7 +718,7 @@ static bitVect *
 getRegsRead (lineNode *line)
 {
   if (!line->aln)
-    line->aln = asmLineNodeFromLineNode (line);
+    line->aln = (asmLineNodeBase *) asmLineNodeFromLineNode (line);
 
   return line->aln->regsRead;
 }
@@ -674,11 +727,48 @@ static bitVect *
 getRegsWritten (lineNode *line)
 {
   if (!line->aln)
-    line->aln = asmLineNodeFromLineNode (line);
+    line->aln = (asmLineNodeBase *) asmLineNodeFromLineNode (line);
 
   return line->aln->regsWritten;
 }
 
+static const char * models[] = 
+{
+  "small",  "small-xstack",  "small-stack-auto",  "small-xstack-auto",
+  "medium", "medium-xstack", "medium-stack-auto", "medium-xstack-auto",
+  "large",  "large-xstack",  "large-stack-auto",  "large-xstack-auto",
+  "huge",   "huge-xstack",   "huge-stack-auto",   "huge-xstack-auto",
+};
+
+static const char *
+get_model (void)
+{
+  int index;
+
+  switch (options.model)
+    {
+    case MODEL_SMALL:
+      index = 0;
+      break;
+    case MODEL_MEDIUM:
+      index = 4;
+      break;
+    case MODEL_LARGE:
+      index = 8;
+      break;
+    case MODEL_HUGE:
+      index = 12;
+      break;
+    default:
+      werror (W_UNKNOWN_MODEL, __FILE__, __LINE__);
+      return "unknown";
+    }
+  if (options.stackAuto)
+    index += 2;
+  if (options.useXstack)
+    index += 1;
+  return models[index];
+}
 
 /** $1 is always the basename.
     $2 is always the output file.
@@ -688,15 +778,17 @@ getRegsWritten (lineNode *line)
 */
 static const char *_linkCmd[] =
 {
-  "aslink", "-nf", "\"$1\"", NULL
+  "sdld", "-nf", "$1", NULL
 };
 
 /* $3 is replaced by assembler.debug_opts resp. port->assembler.plain_opts */
 static const char *_asmCmd[] =
 {
-  "asx8051", "$l", "$3", "\"$1.asm\"", NULL
+  "sdas8051", "$l", "$3", "$2", "$1.asm", NULL
 };
 
+static const char * const _libs[] = { "mcs51", STD_LIB, STD_INT_LIB, STD_LONG_LIB, STD_FP_LIB, NULL, };
+
 /* Globals */
 PORT mcs51_port =
 {
@@ -706,39 +798,44 @@ PORT mcs51_port =
   NULL,                         /* Processor name */
   {
     glue,
-    TRUE,                       /* Emit glue around main */
-    MODEL_SMALL | MODEL_MEDIUM | MODEL_LARGE,
-    MODEL_SMALL
+    TRUE,                       /* glue_up_main: Emit glue around main */
+    MODEL_SMALL | MODEL_MEDIUM | MODEL_LARGE | MODEL_HUGE,
+    MODEL_SMALL,
+    get_model,
   },
-  {
+  {                             /* Assembler */
     _asmCmd,
     NULL,
-    "-plosgffc",                /* Options with debug */
-    "-plosgff",                 /* Options without debug */
+    "-plosgffwy",               /* Options with debug */
+    "-plosgffw",                /* Options without debug */
     0,
     ".asm",
     NULL                        /* no do_assemble function */
   },
-  {
+  {                             /* Linker */
     _linkCmd,
     NULL,
     NULL,
     ".rel",
-    1
+    1,
+    NULL,                       /* crt */
+    _libs,                      /* libs */
   },
-  {
+  {                             /* Peephole optimizer */
     _defaultRules,
     getInstructionSize,
     getRegsRead,
     getRegsWritten,
-    mcs51DeadMove
-  },
-  {
-    /* Sizes: char, short, int, long, ptr, fptr, gptr, bit, float, max */
-    1, 2, 2, 4, 1, 2, 3, 1, 4, 4
+    mcs51DeadMove,
+    NULL,
+    NULL,
+    NULL,
+    NULL,
   },
+  /* Sizes: char, short, int, long, long long, ptr, fptr, gptr, bit, float, max */
+  { 1, 2, 2, 4, 8, 1, 2, 3, 1, 4, 4 },
   /* tags for generic pointers */
-  { 0x00, 0x40, 0x60, 0x80 },		/* far, near, xstack, code */
+  { 0x00, 0x40, 0x60, 0x80 },   /* far, near, xstack, code */
   {
     "XSTK    (PAG,XDATA)",      // xstack_name
     "STACK   (DATA)",           // istack_name
@@ -748,20 +845,23 @@ PORT mcs51_port =
     "PSEG    (PAG,XDATA)",      // pdata_name
     "XSEG    (XDATA)",          // xdata_name
     "BSEG    (BIT)",            // bit_name
-    "RSEG    (DATA)",           // reg_name
+    "RSEG    (ABS,DATA)",       // reg_name
     "GSINIT  (CODE)",           // static_name
     "OSEG    (OVR,DATA)",       // overlay_name
     "GSFINAL (CODE)",           // post_static_name
     "HOME    (CODE)",           // home_name
-    "XISEG   (XDATA)",          // xidata_name - initialized xdata   initialized xdata
+    "XISEG   (XDATA)",          // xidata_name - initialized xdata
     "XINIT   (CODE)",           // xinit_name - a code copy of xiseg
     "CONST   (CODE)",           // const_name - const data (code or not)
     "CABS    (ABS,CODE)",       // cabs_name - const absolute data (code or not)
     "XABS    (ABS,XDATA)",      // xabs_name - absolute xdata/pdata
     "IABS    (ABS,DATA)",       // iabs_name - absolute idata/data
+    NULL,                       // name of segment for initialized variables
+    NULL,                       // name of segment for copies of initialized variables in code space
     NULL,
     NULL,
-    1
+    1,
+    1                           // No fancy alignments supported.
   },
   { _mcs51_genExtraAreas, NULL },
   {
@@ -770,15 +870,11 @@ PORT mcs51_port =
     4,          /* isr_overhead */
     1,          /* call_overhead (2 for return address - 1 for pre-incrementing push */
     1,          /* reent_overhead */
-    0           /* banked_overhead (switch between code banks) */
-  },
-  {
-    /* mcs51 has an 8 bit mul */
-    1, -1
-  },
-  {
-    mcs51_emitDebuggerSymbol
+    1,          /* banked_overhead (switch between code banks) */
+    0           /* sp points directly at last item pushed */
   },
+  { -1, FALSE },
+  { mcs51_emitDebuggerSymbol },
   {
     256,        /* maxCount */
     2,          /* sizeofElement */
@@ -790,12 +886,14 @@ PORT mcs51_port =
   "_",
   _mcs51_init,
   _mcs51_parseOptions,
-  NULL,
+  _mcs51_options,
   NULL,
   _mcs51_finaliseOptions,
   _mcs51_setDefaultOptions,
   mcs51_assignRegisters,
   _mcs51_getRegName,
+  0,
+  _mcs51_rtrackUpdate,
   _mcs51_keywords,
   _mcs51_genAssemblerPreamble,
   NULL,                         /* no genAssemblerEnd */
@@ -804,13 +902,13 @@ PORT mcs51_port =
   _mcs51_genInitStartup,
   _mcs51_reset_regparm,
   _mcs51_regparm,
-  NULL,
-  NULL,
-  NULL,
+  NULL,                         /* process_pragma */
+  NULL,                         /* getMangledFunctionName */
+  _hasNativeMulFor,             /* hasNativeMulFor */
   hasExtBitOp,                  /* hasExtBitOp */
   oclsExpense,                  /* oclsExpense */
-  FALSE,
-  TRUE,                         /* little endian */
+  FALSE,                        /* use_dw_for_init */
+  TRUE,                         /* little_endian */
   0,                            /* leave lt */
   0,                            /* leave gt */
   1,                            /* transform <= to ! > */
@@ -822,6 +920,7 @@ PORT mcs51_port =
   NULL,                         /* no builtin functions */
   GPOINTER,                     /* treat unqualified pointers as "generic" pointers */
   1,                            /* reset labelKey to 1 */
-  1,                            /* globals & local static allowed */
+  1,                            /* globals & local statics allowed */
+  0,                            /* Number of registers handled in the tree-decomposition-based register allocator in SDCCralloc.hpp */
   PORT_MAGIC
 };
